/* The game's drawing surface. Uses active rendering to a JPanel
   with the help of Java 3D's timer.
   See WormP for a version with statistics generation.
*/

import javax.swing.*;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.*;
//import com.sun.j3d.utils.timer.J3DTimer;


public class BugPanel extends JPanel implements Runnable
{
  private static final int PWIDTH = 500;   // size of panel
  private static final int PHEIGHT = 400; 

  // image loader information files
  private static final String IMS_INFO = "imsInfo.txt";

  private Thread animator;           // the thread that performs the animation
  private volatile boolean running = false;   // used to stop the animation thread
  private volatile boolean isPaused = false;

  private long period;                // period between drawing in _milisecs_

  private BugRunner bugTop;

  private BallSprite ball;        // the sprites
  private BatSprite bat;

  private long gameStartTime;   // when the game started
  private int timeSpentInGame;

  // used at game termination
  private volatile boolean gameOver = false;
  private int score = 0;

  // for displaying messages
  private Font msgsFont;
  private FontMetrics metrics;

  // off-screen rendering
  private Graphics dbg; 
  private Image dbImage = null;

  // holds the background image
  private BufferedImage bgImage = null;



  public BugPanel(BugRunner br, long period) {
    bugTop = br;
    this.period = period;

    setDoubleBuffered(false);
    setBackground(Color.black);
    setPreferredSize( new Dimension(PWIDTH, PHEIGHT));

    setFocusable(true);
    requestFocus();    // JPanel ka fokusin, prano ngjarje tastesh

	addKeyListener( new KeyAdapter() {
       public void keyPressed(KeyEvent e)
       { processKey(e);  }
     });

    // ngarko imazhin e prapavis
    ImagesLoader imsLoader = new ImagesLoader(IMS_INFO); 
    bgImage = imsLoader.getImage("bladerunner");

    // krijo sprajtet e lojs
    bat = new BatSprite(PWIDTH, PHEIGHT, imsLoader, 
                                     (int)(period) ); // n ms
    ball = new BallSprite(PWIDTH, PHEIGHT, imsLoader, this, bat); 

    addMouseListener( new MouseAdapter() {
      public void mousePressed(MouseEvent e)
      { testPress(e.getX(), e.getY()); }  // prpuno klikimin
    });

    // vj fontin e mesazheve
    msgsFont = new Font("SansSerif", Font.BOLD, 24);
    metrics = this.getFontMetrics(msgsFont);

  }



  /* Prpunon prfundimin dhe tastet e lojs */
  private void processKey(KeyEvent e) {
    int keyCode = e.getKeyCode();

	// dgjo pr esc, q, end, ctrl-c
    if ((keyCode == KeyEvent.VK_ESCAPE) || (keyCode == KeyEvent.VK_Q) ||
        (keyCode == KeyEvent.VK_END) ||
        ((keyCode == KeyEvent.VK_C) && e.isControlDown()) )
      running = false;
    
    // tastet e lojs
    if (!isPaused && !gameOver) {
      if (keyCode == KeyEvent.VK_LEFT)
        bat.moveLeft(); 
      else if (keyCode == KeyEvent.VK_RIGHT)
        bat.moveRight();
      else if (keyCode == KeyEvent.VK_DOWN)
        bat.stayStill();
    }
  }



  /* Invokohet nga BallSprite pr t sinjalizuar fundin */
  public void gameOver() { 
    int finalTime = 	// ms --> sek
          (int) ((System.currentTimeMillis() - gameStartTime)/1000);
    score = finalTime;   // mund t jet m kompleks!
    gameOver = true;  
  }



  public void addNotify()
  // wait for the JPanel to be added to the JFrame before starting
  { super.addNotify();   // creates the peer
    startGame();         // start the thread
  }


  private void startGame()
  // initialise and start the thread 
  { 
    if (animator == null || !running) {
      animator = new Thread(this);
	  animator.start();
    }
  } // end of startGame()
    

  // ------------- game life cycle methods ------------
  // called by the JFrame's window listener methods


  /* Called when the JFrame is activated / deiconified */
  public synchronized void resumeGame() {
	  isPaused = false;
	  notify();
  } 


  /* Called when the JFrame is deactivated / iconified */
  public void pauseGame() {
	  isPaused = true;
  } 


	/* Invokohet nga JFrame sht duke u mbyllur */
	public synchronized void stopGame() {
		running = false;
		notify();
	}

  // ----------------------------------------------


  /* Shfrytzo mausin pr t kontrolluar shkopin */
  private void testPress(int x, int y) {
	  if (!isPaused && !gameOver)
		  bat.mouseMove(x);
  }


	/* Prsrit: update, render, sleep ashtu q t zgjas period ms */
	public void run() {
		long beforeTime, timeDiff, sleepTime;
		gameStartTime = System.currentTimeMillis();
		beforeTime = gameStartTime;
		running = true;
		while (running) {
			try {
				if (isPaused) {
					synchronized (this) {
						while (isPaused && running)
							wait();
					}
				}
			} catch (InterruptedException e) {
			}
			gameUpdate();
			gameRender();
			paintScreen();
			timeDiff = System.currentTimeMillis() - beforeTime;
			sleepTime = period - timeDiff; // koha e mbetur n iterimin
			if (sleepTime <= 0) // update-render zgjati m tepr
				sleepTime = 5; // megjithat flej pak
			try {
				Thread.sleep(sleepTime); // n ms
			} catch (InterruptedException ex) {
			}
			beforeTime = System.currentTimeMillis();
		}
		System.exit(0);
	}

  private void gameUpdate() {
	  if (!isPaused && !gameOver) {
		  ball.updateSprite();  
		  bat.updateSprite();  
    }
  }


  private void gameRender() {
    if (dbImage == null){
      dbImage = createImage(PWIDTH, PHEIGHT);
      if (dbImage == null) {
        System.out.println("dbImage is null");
        return;
      }
      else
        dbg = dbImage.getGraphics();
    }

    // vizato prapavin: shfrytzo imazhin si ngjyr t zez
    if (bgImage == null) {
      dbg.setColor(Color.black);
      dbg.fillRect (0, 0, PWIDTH, PHEIGHT);
    }
    else
      dbg.drawImage(bgImage, 0, 0, this);

    // vizato elementet e lojs
    ball.drawSprite(dbg); 
    bat.drawSprite(dbg);

    reportStats(dbg);

    if (gameOver)
      gameOverMessage(dbg);
  }


  /* Raporto numrin e topave t kthyer dhe kohn e kaluar n loj */
  private void reportStats(Graphics g) {
    if (!gameOver)    // rrit tajmerim
      timeSpentInGame = 	// ms --> sek
          (int) ((System.currentTimeMillis() - gameStartTime)/1000);

	g.setColor(Color.yellow);
    g.setFont(msgsFont);

    ball.drawBallStats(g, 15, 25);  // sprajti top raporton statistikat e veta
	g.drawString("Time: " + timeSpentInGame + " secs", 15, 50);
   
	g.setColor(Color.black);
  }


  private void gameOverMessage(Graphics g)
  // center the game-over message in the panel
  {
    String msg = "Game Over. Your score: " + score;

	int x = (PWIDTH - metrics.stringWidth(msg))/2; 
	int y = (PHEIGHT - metrics.getHeight())/2;
	g.setColor(Color.red);
    g.setFont(msgsFont);
	g.drawString(msg, x, y);
  }  // end of gameOverMessage()


  private void paintScreen()
  // use active rendering to put the buffered image on-screen
  { 
    Graphics g;
    try {
      g = this.getGraphics();
      if ((g != null) && (dbImage != null))
        g.drawImage(dbImage, 0, 0, null);
      // Sync the display on some systems.
      // (on Linux, this fixes event queue problems)
      Toolkit.getDefaultToolkit().sync();

      g.dispose();
    }
    catch (Exception e)
    { System.out.println("Graphics context error: " + e);  }
  } // end of paintScreen()


}  // end of BugPanel class
